/* noteLED
 *
 * (C) 2012 by morecat_lab
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbdrv.h"
#include "oddebug.h"

/*
  PIN ASSIGN

  PB0 : (ONBOARD LED)
  PB1 : C
  PB2 : C#
  PB3 : D
  PB4 : D#
  PB5 : E
  PB6 : F
  PB7 : F#

  PD0 : G
  PD1 : G#
  PD2 : (USB D+)
  PD3 : (USB D-)
  PD4 : A
  PD5 : A#
  PD6 : B


 */

#define setPortBit0(port, bit) *(port) &= ~_BV(bit);
#define setPortBit1(port, bit) *(port) |= _BV(bit);

#define setPortBit(port, bit, val) { \
    if ((val) == 0) { setPortBit0((port), (bit)); } else	\
      { setPortBit1((port), (bit)); }				\
}
#define getPortBit(port, bit) ((*(port) & _BV(bit)) >> (bit))

extern void parseMidiMessage(uchar *, uchar);

/* debugging LED */
#define LED_OFF  { PORTB &= ~0x01; }
#define LED_ON   { PORTB |=  0x01; }
#define FLIP_LED { PORTB ^=  0x01; }

#define HW_CDC_BULK_OUT_SIZE     8
#define HW_CDC_BULK_IN_SIZE      8

enum {
    SEND_ENCAPSULATED_COMMAND = 0,
    GET_ENCAPSULATED_RESPONSE,
    SET_COMM_FEATURE,
    GET_COMM_FEATURE,
    CLEAR_COMM_FEATURE,
    SET_LINE_CODING = 0x20,
    GET_LINE_CODING,
    SET_CONTROL_LINE_STATE,
    SEND_BREAK
};

// This deviceDescrMIDI[] is based on 
// http://www.usb.org/developers/devclass_docs/midi10.pdf
// Appendix B. Example: Simple MIDI Adapter (Informative)

// B.1 Device Descriptor
static PROGMEM char deviceDescrMIDI[] = {	/* USB device descriptor */
  18,			/* sizeof(usbDescriptorDevice): length of descriptor in bytes */
  USBDESCR_DEVICE,	/* descriptor type */
  0x10, 0x01,		/* USB version supported */
  0,			/* device class: defined at interface level */
  0,			/* subclass */
  0,			/* protocol */
  8,			/* max packet size */
  USB_CFG_VENDOR_ID,	/* 2 bytes */
  USB_CFG_DEVICE_ID,	/* 2 bytes */
  USB_CFG_DEVICE_VERSION,	/* 2 bytes */
  1,			/* manufacturer string index */
  2,			/* product string index */
  0,			/* serial number string index */
  1,			/* number of configurations */
};

// B.2 Configuration Descriptor
static PROGMEM char configDescrMIDI[] = { /* USB configuration descriptor */
  9,	   /* sizeof(usbDescrConfig): length of descriptor in bytes */
  USBDESCR_CONFIG,		/* descriptor type */
  101, 0,			/* total length of data returned (including inlined descriptors) */
  2,		      /* number of interfaces in this configuration */
  1,				/* index of this configuration */
  0,				/* configuration name string index */
#if USB_CFG_IS_SELF_POWERED
    (1 << 7) | USBATTR_SELFPOWER,       /* attributes */
#else
    (1 << 7),                           /* attributes */
#endif
  USB_CFG_MAX_BUS_POWER / 2,	/* max USB current in 2mA units */

// B.3 AudioControl Interface Descriptors
// The AudioControl interface describes the device structure (audio function topology) 
// and is used to manipulate the Audio Controls. This device has no audio function 
// incorporated. However, the AudioControl interface is mandatory and therefore both 
// the standard AC interface descriptor and the classspecific AC interface descriptor 
// must be present. The class-specific AC interface descriptor only contains the header 
// descriptor.

// B.3.1 Standard AC Interface Descriptor
// The AudioControl interface has no dedicated endpoints associated with it. It uses the 
// default pipe (endpoint 0) for all communication purposes. Class-specific AudioControl 
// Requests are sent using the default pipe. There is no Status Interrupt endpoint provided.
  /* descriptor follows inline: */
  9,			/* sizeof(usbDescrInterface): length of descriptor in bytes */
  USBDESCR_INTERFACE,	/* descriptor type */
  0,			/* index of this interface */
  0,			/* alternate setting for this interface */
  0,			/* endpoints excl 0: number of endpoint descriptors to follow */
  1,			/* */
  1,			/* */
  0,			/* */
  0,			/* string index for interface */

// B.3.2 Class-specific AC Interface Descriptor
// The Class-specific AC interface descriptor is always headed by a Header descriptor 
// that contains general information about the AudioControl interface. It contains all 
// the pointers needed to describe the Audio Interface Collection, associated with the 
// described audio function. Only the Header descriptor is present in this device 
// because it does not contain any audio functionality as such.
  /* descriptor follows inline: */
  9,			/* sizeof(usbDescrCDC_HeaderFn): length of descriptor in bytes */
  36,			/* descriptor type */
  1,			/* header functional descriptor */
  0x0, 0x01,		/* bcdADC */
  9, 0,			/* wTotalLength */
  1,			/* */
  1,			/* */

// B.4 MIDIStreaming Interface Descriptors

// B.4.1 Standard MS Interface Descriptor
  /* descriptor follows inline: */
  9,			/* length of descriptor in bytes */
  USBDESCR_INTERFACE,	/* descriptor type */
  1,			/* index of this interface */
  0,			/* alternate setting for this interface */
  2,			/* endpoints excl 0: number of endpoint descriptors to follow */
  1,			/* AUDIO */
  3,			/* MS */
  0,			/* unused */
  0,			/* string index for interface */

// B.4.2 Class-specific MS Interface Descriptor
  /* descriptor follows inline: */
  7,			/* length of descriptor in bytes */
  36,			/* descriptor type */
  1,			/* header functional descriptor */
  0x0, 0x01,		/* bcdADC */
  65, 0,			/* wTotalLength */

// B.4.3 MIDI IN Jack Descriptor
  /* descriptor follows inline: */
  6,			/* bLength */
  36,			/* descriptor type */
  2,			/* MIDI_IN_JACK desc subtype */
  1,			/* EMBEDDED bJackType */
  1,			/* bJackID */
  0,			/* iJack */

  /* descriptor follows inline: */
  6,			/* bLength */
  36,			/* descriptor type */
  2,			/* MIDI_IN_JACK desc subtype */
  2,			/* EXTERNAL bJackType */
  2,			/* bJackID */
  0,			/* iJack */

//B.4.4 MIDI OUT Jack Descriptor
  /* descriptor follows inline: */
  9,			/* length of descriptor in bytes */
  36,			/* descriptor type */
  3,			/* MIDI_OUT_JACK descriptor */
  1,			/* EMBEDDED bJackType */
  3,			/* bJackID */
  1,			/* No of input pins */
  2,			/* BaSourceID */
  1,			/* BaSourcePin */
  0,			/* iJack */

  /* descriptor follows inline: */
  9,			/* bLength of descriptor in bytes */
  36,			/* bDescriptorType */
  3,			/* MIDI_OUT_JACK bDescriptorSubtype */
  2,			/* EXTERNAL bJackType */
  4,			/* bJackID */
  1,			/* bNrInputPins */
  1,			/* baSourceID (0) */
  1,			/* baSourcePin (0) */
  0,			/* iJack */


// B.5 Bulk OUT Endpoint Descriptors

//B.5.1 Standard Bulk OUT Endpoint Descriptor
  /* descriptor follows inline: */
  9,			/* bLenght */
  USBDESCR_ENDPOINT,	/* bDescriptorType = endpoint */
  0x1,			/* bEndpointAddress OUT endpoint number 1 */
  3,			/* bmAttributes: 2:Bulk, 3:Interrupt endpoint */
  8, 0,			/* wMaxPacketSize */
  10,			/* bIntervall in ms */
  0,			/* bRefresh */
  0,			/* bSyncAddress */

// B.5.2 Class-specific MS Bulk OUT Endpoint Descriptor
  /* descriptor follows inline: */
  5,			/* bLength of descriptor in bytes */
  37,			/* bDescriptorType */
  1,			/* bDescriptorSubtype */
  1,			/* bNumEmbMIDIJack  */
  1,			/* baAssocJackID (0) */


//B.6 Bulk IN Endpoint Descriptors

//B.6.1 Standard Bulk IN Endpoint Descriptor
  /* descriptor follows inline: */
  9,			/* bLenght */
  USBDESCR_ENDPOINT,	/* bDescriptorType = endpoint */
  0x81,			/* bEndpointAddress IN endpoint number 1 */
  3,			/* bmAttributes: 2: Bulk, 3: Interrupt endpoint */
  8, 0,			/* wMaxPacketSize */
  10,			/* bIntervall in ms */
  0,			/* bRefresh */
  0,			/* bSyncAddress */

// B.6.2 Class-specific MS Bulk IN Endpoint Descriptor
  /* descriptor follows inline: */
  5,			/* bLength of descriptor in bytes */
  37,			/* bDescriptorType */
  1,			/* bDescriptorSubtype */
  1,			/* bNumEmbMIDIJack (0) */
  3,			/* baAssocJackID (0) */
};


uchar usbFunctionDescriptor(usbRequest_t * rq) {
  if (rq->wValue.bytes[1] == USBDESCR_DEVICE) {
    usbMsgPtr = (uchar *) deviceDescrMIDI;
    return sizeof(deviceDescrMIDI);
  } else {		/* must be config descriptor */
    usbMsgPtr = (uchar *) configDescrMIDI;
    return sizeof(configDescrMIDI);
  }
}

/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

static uchar    sendEmptyFrame;
// static uchar    modeBuffer[7];

uchar usbFunctionSetup(uchar data[8]) {
  usbRequest_t    *rq = (void *)data;

  if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
    
    /*  Prepare bulk-in endpoint to respond to early termination   */
    if ((rq->bmRequestType & USBRQ_DIR_MASK) ==
	USBRQ_DIR_HOST_TO_DEVICE)
      sendEmptyFrame = 1;
  }
  return 0xff;
}

/*---------------------------------------------------------------------------*/
/* usbFunctionRead                                                           */
/*---------------------------------------------------------------------------*/

uchar usbFunctionRead( uchar *data, uchar len ) {
	data[0] = 0;
	data[1] = 0;
	data[2] = 0;
	data[3] = 0;
	data[4] = 0;
	data[5] = 0;
	data[6] = 0;
    return 7;
}

/*---------------------------------------------------------------------------*/
/* usbFunctionWrite                                                          */
/*---------------------------------------------------------------------------*/

uchar usbFunctionWrite(uchar * data, uchar len) {
  return 1;
}
/*---------------------------------------------------------------------------*/
/* usbFunctionWriteOut                                                       */
/*                                                                           */
/* this Function is called if a MIDI Out message (from PC) arrives.          */
/*                                                                           */
/*---------------------------------------------------------------------------*/

void usbFunctionWriteOut(uchar * data, uchar len) {
  /* flipPortBit(&PORTB, 5);		*/
  parseMidiMessage(data, len);
}

/*----------------------------------------------- */
/* Key Event処理 */

/* keyEventのQueue */
struct keyEvent {
  uchar key;
  //  uchar onOff;
};

#define KEY_OFF 0
#define KEY_ON 1
#define MAX_KEY_EVENT 32
#define KEY_EVENT_QUEUE_MASK (MAX_KEY_EVENT-1)
#define KEY_COUNT_TH 10

struct keyEvent keyEventQueue[MAX_KEY_EVENT];
uchar urptr, uwptr, iwptr;

static inline uchar keyEventsFree (struct keyEvent **ptr) {
  return (MAX_KEY_EVENT - uwptr - 1);
}

void keyEventAppend(uchar key, uchar onOff) {
  uchar iwnxt;
  iwnxt = (iwptr +1 ) & KEY_EVENT_QUEUE_MASK;
  if (iwnxt == urptr )
    return;			/* buffer overflow */
  keyEventQueue[iwptr].key = key | (onOff << 7);
  iwptr = iwnxt;
}

static inline uchar keyEventAvailable(struct keyEvent **ptr) {
  *ptr = &keyEventQueue[urptr];
  if(iwptr >= urptr) {
    return iwptr - urptr;
  } else {  /* buffer end is between read and write pointer, return continuous range */
    return MAX_KEY_EVENT - urptr;
  }
}

/* The following function must be called after uartRxBytesAvailable() to
 * remove the bytes from the receiver buffer.
 */
static inline void  keyEventDidReadEvents(uchar numEvents) {
  urptr = (urptr + numEvents) & KEY_EVENT_QUEUE_MASK;
}

static inline void keyEventQueueClear() {
  urptr = uwptr = 0;
}

/* PORT MAP TABLE */
struct portInfo_t {
  uchar note_no;
  volatile uchar *port;
  uchar bit;
  uchar status;
  uchar keyCountButton;
} portInfoOut[] = {
  {60, &PORTB, 1, 0, 0},		/* IX:00 C3 */
  {61, &PORTB, 2, 0, 0},		/* IX:01 C#3 */
  {62, &PORTB, 3, 0, 0},		/* IX:02 D3 */
  {63, &PORTB, 4, 0, 0},		/* IX:03 D#3 */
  {64, &PORTB, 5, 0, 0},		/* IX:04 E3 */
  {65, &PORTB, 6, 0, 0},		/* IX:05 F3 */
  {66, &PORTB, 7, 0, 0},		/* IX:06 F#3 */
  {67, &PORTD, 0, 0, 0},		/* IX:07 G3 */
  {68, &PORTD, 1, 0, 0},		/* IX:08 G#3 */
  {69, &PORTD, 4, 0, 0},		/* IX:09 A3 */
  {70, &PORTD, 5, 0, 0},		/* IX:10 A#3 */
  {71, &PORTD, 6, 0, 0},		/* IX:11 B3 */
};

#define MAX_KEYS 12

#define PORTD_DDR (0b01110011)

inline void NoteON(uchar note, uchar vol) {
  uchar i;
  for (i = 0 ; i < MAX_KEYS ; i++) {
    if (portInfoOut[i].note_no == note) {
      portInfoOut[i].status = 1;
      setPortBit1(portInfoOut[i].port, portInfoOut[i].bit);
    }
  }
}

/* ノートOFF処理 */

inline void NoteOFF(uchar note) {
  uchar i;
  for (i = 0 ; i < MAX_KEYS ; i++) {
    if (portInfoOut[i].note_no == note) {
      portInfoOut[i].status = 0;
      setPortBit0(portInfoOut[i].port, portInfoOut[i].bit);
    }
  }
}

inline void allNoteOFF() {
  uchar i;
  for (i = 0 ; i < MAX_KEYS ; i++) {
    if (portInfoOut[i].status == 1) {
      portInfoOut[i].status = 0;
      setPortBit0(portInfoOut[i].port, portInfoOut[i].bit);
    }
  }
}

/*----------------------------------------------- */
/* メッセージのパース処理*/
/*----------------------------------------------- */
void parseMidiMessage(uchar *data, uchar len) {
  uchar cin = (*data) & 0x0f;	/* CABLE NOを無視する */
  uchar note = (*(data+2)) & 0x7f;
  uchar vol  = (*(data+3)) & 0x7f;

  if (((*(data+1)) & 0x0f) != 0) /* only channel 1 (=0) */
    return;

  switch(cin) {

    /* NOTE OFF */
  case 8:
    NoteOFF(note);
    LED_OFF;
    break;

    /* NOTE ON */
  case 9:
    if( vol == 0){
      NoteOFF(note);
      LED_OFF;
    } else {
      NoteON(note, vol);
      LED_ON;
    }
    /* CONTROL CHANGE */
  case 11:			
    if ((note == 120) || (note == 123)) {
      /* all note off */
      allNoteOFF();
    }
    break;

    //  case 0:			/* MISC FUNCTION CODE */
    //  case 1:			/* CABLE EVENT */
    //  case 2:			/* TWO-BYTES SYSTEM COMMON */
    //  case 3:			/* THREE-BYTES SYSTEM COMMON */
    //  case 4:			/* SYS EX START */
    //  case 5:			/* SINGLE BYTE SYSTEM COMMON */
    //  case 6:			/* SYSTEM EX end (with 2 byte) */
    //  case 7:			/* SYSTEM EX end (with 3 byte) */
    //  case 10:		/* POLY KEY PRESS */
    //  case 12:		/* PROGRAM CHANGE */
    //  case 13:		/* CHANNEL PRESSURE */
    //  case 14:		/* PITCH BEND */
    //  case 15:			/* SINGLE BYTE */
    //    break;
  }

  if (len > 4) {
    parseMidiMessage(data+4, len-4);
  }
}

void hardwareInit(void) {
  uchar i, j;

  /* activate pull-ups except on USB lines */
  USB_CFG_IOPORT =
    (uchar) ~ ((1 << USB_CFG_DMINUS_BIT) |
	       (1 << USB_CFG_DPLUS_BIT) | PORTD_DDR) ;
  /* all pins input except USB (-> USB reset) */
#ifdef USB_CFG_PULLUP_IOPORT
  /* use usbDeviceConnect()/usbDeviceDisconnect() if available */
  USBDDR = 0 | PORTD_DDR;	/* we do RESET by deactivating pullup */
  usbDeviceDisconnect();
#else
  USBDDR = (1 << USB_CFG_DMINUS_BIT) | (1 << USB_CFG_DPLUS_BIT) | PORTD_DDR ;
  /* for output (LED, TxD) */
#endif

  j = 0;
  while (--j) {		/* USB Reset by device only required on Watchdog Reset */
    i = 0;
    while (--i);	/* delay >10ms for USB reset */
  }
#ifdef USB_CFG_PULLUP_IOPORT
  usbDeviceConnect();
#else
  USBDDR = 0 | PORTD_DDR;		/*  remove USB reset condition */
#endif

  DDRB  =  0b11111111;
  PORTB =  0b00000000;			/* All LOW, LED OFF */
  PORTD |= 0b00000000;			/* All LOW */
}

int main(void) {
  /* WDTを作動させる */
  wdt_enable(WDTO_1S);
  odDebugInit();
  hardwareInit();
  usbInit();
  
  sendEmptyFrame = 0;
  sei();

  for(;;){
    wdt_reset();
    usbPoll();
  } /* main event loop */
}

/* EOF */
